// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (C) 1993- 1996, 1995  Microsoft Corporation.  All Rights Reserved.
//
//  MODULE:   service.c
//
//  PURPOSE:  Implements functions required by all services
//            windows.
//
//  FUNCTIONS:
//    main(int argc, char **argv);
//    service_ctrl(DWORD dwCtrlCode);
//    service_main(DWORD dwArgc, LPTSTR *lpszArgv);
//    CmdInstallService();
//    CmdRemoveService();
//    CmdDebugService(int argc, char **argv);
//    ControlHandler ( DWORD dwCtrlType );
//    GetLastErrorText( LPTSTR lpszBuf, DWORD dwSize );
//
//  COMMENTS:
//
//  AUTHOR: Craig Link - Microsoft Developer Support
//


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>

#include "service.h"

#include "../../../Common/error.h"
#include "common.h"



// internal variables
SERVICE_STATUS          ssStatus;       // current status of the service
SERVICE_STATUS_HANDLE   sshStatusHandle;
DWORD                   dwErr = 0;
BOOL                    g_Debug = FALSE;
int                     g_LogLevel = 1;

// internal function prototypes
VOID WINAPI service_ctrl(DWORD dwCtrlCode);
VOID WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv);
VOID CmdInstallService();
VOID CmdRemoveService();
VOID CmdDebugService(int argc, char **argv);
BOOL WINAPI ControlHandler ( DWORD dwCtrlType );

//
//  FUNCTION: main
//
//  PURPOSE: entrypoint for service
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    main() either performs the command line task, or
//    call StartServiceCtrlDispatcher to register the
//    main service thread.  When the this call returns,
//    the service has stopped, so exit.
//
void main(int argc, char **argv)
{
  SERVICE_TABLE_ENTRY dispatchTable[] =
  {
    { TEXT(SZSERVICENAME), (LPSERVICE_MAIN_FUNCTION)service_main },
    { NULL, NULL }
  };
  
  if ( (argc > 1) &&
    ((*argv[1] == '-') || (*argv[1] == '/')) )
  {
    if ( _stricmp( "install", argv[1]+1 ) == 0 )
    {
      CmdInstallService();
    }
    else if ( _stricmp( "remove", argv[1]+1 ) == 0 )
    {
      CmdRemoveService();
    }
    else if ( _stricmp( "debug", argv[1]+1 ) == 0 )
    {
      g_Debug = TRUE;
      CmdDebugService(argc, argv);
    }
    else
    {
      goto dispatch;
    }
    exit(0);
  }
  
  // if it doesn't match any of the above parameters
  // the service control manager may be starting the service
  // so we must call StartServiceCtrlDispatcher
dispatch:
  // this is just to be friendly
  printf("\n");
  printf("NetSaint Server v1.0\n");
  printf("Written By Jan Christian Kaldestad\n");		
  printf("\n");
  printf("Usage: %s [option]\n",argv[0]);
  printf("\n");
  printf("Options:\n"); 
  printf("\n");
  printf("  -install     to install as a service\n");
  printf("  -remove      to remove service\n");
  printf("  -debug       to run as a console app for debugging\n");
  printf("\n");
  printf("\n");
  printf("StartServiceCtrlDispatcher() being called.\n" );
  printf("This may take several seconds.  Please wait...\n" );
  
  if (!StartServiceCtrlDispatcher(dispatchTable))
  {
    //AddToMessageLog(TEXT("StartServiceCtrlDispatcher failed."), EVENTLOG_ERROR_TYPE);
  }
  
  printf("\n");
}



//
//  FUNCTION: service_main
//
//  PURPOSE: To perform actual initialization of the service
//
//  PARAMETERS:
//    dwArgc   - number of command line arguments
//    lpszArgv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//    This routine performs the service initialization and then calls
//    the user defined ServiceStart() routine to perform majority
//    of the work.
//
void WINAPI service_main(DWORD dwArgc, LPTSTR *lpszArgv)
{
  
  // register our service control handler:
  //
  sshStatusHandle = RegisterServiceCtrlHandler( TEXT(SZSERVICENAME), service_ctrl);
  
  if (!sshStatusHandle)
    goto cleanup;
  
  // SERVICE_STATUS members that don't change in example
  //
  ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
  ssStatus.dwServiceSpecificExitCode = 0;
  
  
  // report the status to the service control manager.
  //
  if (!ReportStatusToSCMgr(
    SERVICE_START_PENDING, // service state
    NO_ERROR,              // exit code
    3000))                 // wait hint
    goto cleanup;
  
  
  ServiceStart( dwArgc, lpszArgv );
  
cleanup:
  
  // try to report the stopped status to the service control manager.
  //
  if (sshStatusHandle)
  {
    (VOID)ReportStatusToSCMgr(
    SERVICE_STOPPED,
    dwErr,
    0);
  }
  
  return;
}



//
//  FUNCTION: service_ctrl
//
//  PURPOSE: This function is called by the SCM whenever
//           ControlService() is called on this service.
//
//  PARAMETERS:
//    dwCtrlCode - type of control requested
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
VOID WINAPI service_ctrl(DWORD dwCtrlCode)
{
  // Handle the requested control code.
  //
  switch(dwCtrlCode)
  {
    // Stop the service.
    //
  case SERVICE_CONTROL_STOP:
    ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 0);
    ServiceStop();
    return;
    
    // Update the service status.
    //
  case SERVICE_CONTROL_INTERROGATE:
    break;
    
    // invalid control code
    //
  default:
    break;
    
  }
  
  ReportStatusToSCMgr(ssStatus.dwCurrentState, NO_ERROR, 0);
  
}



//
//  FUNCTION: ReportStatusToSCMgr()
//
//  PURPOSE: Sets the current status of the service and
//           reports it to the Service Control Manager
//
//  PARAMETERS:
//    dwCurrentState - the state of the service
//    dwWin32ExitCode - error code to report
//    dwWaitHint - worst case estimate to next checkpoint
//
//  RETURN VALUE:
//    TRUE  - success
//    FALSE - failure
//
//  COMMENTS:
//
BOOL ReportStatusToSCMgr(DWORD dwCurrentState,
                         DWORD dwWin32ExitCode,
                         DWORD dwWaitHint)
{
  static DWORD dwCheckPoint = 1;
  BOOL fResult = TRUE;
  
  
  if ( !g_Debug ) // when debugging we don't report to the SCM
  {
    if (dwCurrentState == SERVICE_START_PENDING)
      ssStatus.dwControlsAccepted = 0;
    else
      ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    
    ssStatus.dwCurrentState = dwCurrentState;
    ssStatus.dwWin32ExitCode = dwWin32ExitCode;
    ssStatus.dwWaitHint = dwWaitHint;
    
    if ( ( dwCurrentState == SERVICE_RUNNING ) ||
      ( dwCurrentState == SERVICE_STOPPED ) )
      ssStatus.dwCheckPoint = 0;
    else
      ssStatus.dwCheckPoint = dwCheckPoint++;
    
    
    // Report the status of the service to the service control manager.
    //
    if (!(fResult = SetServiceStatus( sshStatusHandle, &ssStatus))) {
      ;//AddToMessageLog(TEXT("SetServiceStatus"), EVENTLOG_ERROR_TYPE);
    }
  }
  return fResult;
}


///////////////////////////////////////////////////////////////////
//
//  The following code handles service installation and removal
//


//
//  FUNCTION: CmdInstallService()
//
//  PURPOSE: Installs the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdInstallService()
{
  SC_HANDLE   schService;
  SC_HANDLE   schSCManager;
  
  TCHAR szPath[512];
  
  if ( GetModuleFileName( NULL, szPath, 512 ) == 0 )
  {
    _tprintf(TEXT("Unable to install %s - %s\n"), TEXT(SZSERVICEDISPLAYNAME), GetLastErrorText());
    return;
  }
  
  schSCManager = OpenSCManager(
    NULL,                   // machine (NULL == local)
    NULL,                   // database (NULL == default)
    SC_MANAGER_ALL_ACCESS   // access required
    );
  if ( schSCManager )
  {
    schService = CreateService(
      schSCManager,               // SCManager database
      TEXT(SZSERVICENAME),        // name of service
      TEXT(SZSERVICEDISPLAYNAME), // name to display
      SERVICE_ALL_ACCESS,         // desired access
      SERVICE_WIN32_OWN_PROCESS,  // service type
      SERVICE_DEMAND_START,       // start type
      SERVICE_ERROR_NORMAL,       // error control type
      szPath,                     // service's binary
      NULL,                       // no load ordering group
      NULL,                       // no tag identifier
      TEXT(SZDEPENDENCIES),       // dependencies
      NULL,                       // LocalSystem account
      NULL);                      // no password
    
    if ( schService )
    {
      _tprintf(TEXT("%s installed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
      CloseServiceHandle(schService);
    }
    else
    {
      _tprintf(TEXT("CreateService failed - %s\n"), GetLastErrorText());
    }
    
    CloseServiceHandle(schSCManager);
  }
  else
    _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText());
}



//
//  FUNCTION: CmdRemoveService()
//
//  PURPOSE: Stops and removes the service
//
//  PARAMETERS:
//    none
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdRemoveService()
{
  SC_HANDLE   schService;
  SC_HANDLE   schSCManager;
  
  schSCManager = OpenSCManager(
    NULL,                   // machine (NULL == local)
    NULL,                   // database (NULL == default)
    SC_MANAGER_ALL_ACCESS   // access required
    );
  if ( schSCManager )
  {
    schService = OpenService(schSCManager, TEXT(SZSERVICENAME), SERVICE_ALL_ACCESS);
    
    if (schService)
    {
      // try to stop the service
      if ( ControlService( schService, SERVICE_CONTROL_STOP, &ssStatus ) )
      {
        _tprintf(TEXT("Stopping %s."), TEXT(SZSERVICEDISPLAYNAME));
        Sleep( 1000 );
        
        while( QueryServiceStatus( schService, &ssStatus ) )
        {
          if ( ssStatus.dwCurrentState == SERVICE_STOP_PENDING )
          {
            _tprintf(TEXT("."));
            Sleep( 1000 );
          }
          else
            break;
        }
        
        if ( ssStatus.dwCurrentState == SERVICE_STOPPED )
          _tprintf(TEXT("\n%s stopped.\n"), TEXT(SZSERVICEDISPLAYNAME) );
        else
          _tprintf(TEXT("\n%s failed to stop.\n"), TEXT(SZSERVICEDISPLAYNAME) );
        
      }
      
      // now remove the service
      if( DeleteService(schService) )
        _tprintf(TEXT("%s removed.\n"), TEXT(SZSERVICEDISPLAYNAME) );
      else
        _tprintf(TEXT("DeleteService failed - %s\n"), GetLastErrorText());
      
      
      CloseServiceHandle(schService);
    }
    else
      _tprintf(TEXT("OpenService failed - %s\n"), GetLastErrorText());
    
    CloseServiceHandle(schSCManager);
  }
  else
    _tprintf(TEXT("OpenSCManager failed - %s\n"), GetLastErrorText());
}




///////////////////////////////////////////////////////////////////
//
//  The following code is for running the service as a console app
//


//
//  FUNCTION: CmdDebugService(int argc, char ** argv)
//
//  PURPOSE: Runs the service as a console application
//
//  PARAMETERS:
//    argc - number of command line arguments
//    argv - array of command line arguments
//
//  RETURN VALUE:
//    none
//
//  COMMENTS:
//
void CmdDebugService(int argc, char ** argv)
{
  DWORD dwArgc;
  LPTSTR *lpszArgv;
  
#ifdef UNICODE
  lpszArgv = CommandLineToArgvW(GetCommandLineW(), &(dwArgc) );
#else
  dwArgc   = (DWORD) argc;
  lpszArgv = argv;
#endif
  
  SetConsoleCtrlHandler( ControlHandler, TRUE );
  
  ServiceStart( dwArgc, lpszArgv );

  printToLog(DEBUG_LEVEL, TEXT("Debugging %s.\n"), TEXT(SZSERVICEDISPLAYNAME));
}


//
//  FUNCTION: ControlHandler ( DWORD dwCtrlType )
//
//  PURPOSE: Handled console control events
//
//  PARAMETERS:
//    dwCtrlType - type of control event
//
//  RETURN VALUE:
//    True - handled
//    False - unhandled
//
//  COMMENTS:
//
BOOL WINAPI ControlHandler ( DWORD dwCtrlType )
{
  switch( dwCtrlType )
  {
  case CTRL_BREAK_EVENT:  // use Ctrl+C or Ctrl+Break to simulate
  case CTRL_C_EVENT:      // SERVICE_CONTROL_STOP in debug mode
    _tprintf(TEXT("Stopping %s.\n"), TEXT(SZSERVICEDISPLAYNAME));
    ServiceStop();
    return TRUE;
    break;
    
  }
  return FALSE;
}

